class NotFound(Exception): pass
"""Used to signify no value was found"""

def value_at(data, path, chain=None):
    """Return the value at this path"""
    if not chain:
        chain = []

    if not path:
        return chain, data

    elif path and not isinstance(data, dict):
        raise NotFound

    keys = list(reversed(sorted(data.keys())))

    if path in keys:
        return chain + [path], data[path]

    for key in keys:
        if path.startswith("{0}.".format(key)):
            try:
                return value_at(data[key], path[len(key)+1:], chain=chain + [key])
            except NotFound:
                pass

    raise NotFound

def without_prefix(path, prefix=""):
    """
    Remove the prefix from a path

    If the prefix isn't on this path, just return the path itself

    If the path is the prefix, return the prefix
    """
    if not prefix or not path or prefix == path:
        return path

    if path.startswith("{0}.".format(prefix)):
        return path[len(prefix)+1:]

    return path

def prefixed_path_list(path, prefix=None):
    """Return the prefixed version of this path as a list"""
    if not prefix:
        return path

    if prefix is None:
        prefix = []

    return prefix + path

def prefixed_path_string(path, prefix=""):
    """Return the prefixed version of this string"""
    while path and path.startswith("."):
        path = path[1:]

    while path and path.endswith("."):
        path = path[:-1]

    while prefix and prefix.startswith("."):
        prefix = prefix[1:]

    while prefix and prefix.endswith("."):
        prefix = prefix[:-1]

    result = []
    if prefix:
        result.append(prefix)
    if path:
        result.append(path)
    return '.'.join(result)

def dot_joiner(lst):
    """Join list of strings with a single dot in between each"""
    result = []
    for part in lst:
        while part and part.startswith("."):
            part = part[1:]

        while part and part.endswith("."):
            part = part[:-1]

        if part:
            result.append(part)

    return '.'.join(result)

def make_dict(first, rest, data):
    """Make a dictionary from a list of keys"""
    last = first
    result = {first: data}
    current = result
    for part in rest:
        current[last] = {}
        current = current[last]
        current[part] = data
        last = part

    return result




from option_merge.helper import prefixed_path_list, prefixed_path_string, without_prefix, dot_joiner, make_dict
from option_merge.storage import Storage

from delfick_error import DelfickError
from collections import Mapping
import six

class BadPrefix(DelfickError): pass
"""Used to say a prefix value is not a dictionary"""

class KeyValuePairsConverter(object):
    """Converts a list of key,value pairs to a dictionary"""
    def __init__(self, pairs, source=None):
        self.pairs = pairs
        self.source = source

    def convert(self):
        """Return us a MergedOptions from our pairs"""
        return MergedOptions().using(*[make_dict(key[0], key[1:], value) for key, value in self.pairs], source=self.source)

class AttributesConverter(object):
    """Converts an object with particular attributes to a dictionary"""
    def __init__(self, obj, attributes=None, include_underlined=False, lift=None, ignoreable_values=None, source=None):
        self.obj = obj
        self.lift = lift
        self.source = source
        self.attributes = attributes
        self.ignoreable_values = ignoreable_values
        self.include_underlined = include_underlined

        if isinstance(self.lift, six.string_types):
            self.lift = [self.lift]

    def convert(self):
        """Return us a MergedOptions from our attributes"""
        options = MergedOptions()
        attributes = self.attributes
        if not attributes:
            attributes = dir(self.obj)
            if not self.include_underlined:
                attributes = [attr for attr in attributes if not attr.startswith("_")]

        for attr in attributes:
            if hasattr(self.obj, attr):
                val = getattr(self.obj, attr)
                if not self.ignoreable_values or val not in self.ignoreable_values:
                    options[attr] = val

        if self.lift:
            lifted = MergedOptions()
            lifted.storage.add(self.lift, options)
            return lifted
        else:
            return options

class ConverterProperty(object):
    """Creates a Property for accessing a converter"""
    def __init__(self, converter):
        self.converter = converter

    def __get__(self, obj=None, owner=None):
        return lambda *args, **kwargs: self.converter(*args, **kwargs).convert()

class MergedOptions(dict, Mapping):
    """
    Wrapper around multiple dictionaries to behave as one.

    Usage::

        options = MergedOptions.using(options1, options2, source="SomePlace")

    Is equivalent to::

        options = MergedOptions()
        options.update(options1, source="SomePlace")
        options.update(options2, source=SomePlace")

    The later an option is added, the more influence it has.
    i.e. when a key is accessed, later options are looked at first.

    When you delete a key, it removes it from the first dictionary it can find.
    This means a key can change value when deleted rather than disappearing altogether

    It will also merge deeply.

    So::

        options1 = {'a':{'b':1, 'c':3}, 'b':5}
        options2 = {'a':{'b':4'}, 'd':7}
        merged = MergedOptions.using(options1, options2)
        merged['a'] == MergedOptions(prefix='a', <same_options>, <same_overrides>)
        merged['a']['b'] == 4
        merged['a']['c'] == 3
        merged['d'] == 7
    """

    Attributes = ConverterProperty(AttributesConverter)
    KeyValuePairs = ConverterProperty(KeyValuePairsConverter)

    def __init__(self, prefix=None, storage=None):
        self.prefix_list = prefix
        if not self.prefix_list:
            self.prefix_list = []
        if isinstance(self.prefix_list, six.string_types):
            self.prefix_list = [self.prefix_list]
        self.prefix_string = dot_joiner(self.prefix_list)

        self.storage = storage
        if self.storage is None:
            self.storage = Storage()

    @classmethod
    def using(cls, *options, **kwargs):
        """Convenience for calling update multiple times"""
        source = kwargs.get("source")
        merged = cls()
        for opts in options:
            merged.update(opts, source=source)
        return merged

    def update(self, options, source=None):
        """Add new options"""
        if options is None: return
        self.storage.add(self.prefix_list, options, source=source)

    def __getitem__(self, path):
        """
        Access some path

        Return the first value it comes across
        Raise KeyError if nothing has the specified key
        """
        for val in self.values_for(path):
            if isinstance(val, dict):
                return self.prefixed(path)
            else:
                return val
        raise KeyError(path)

    def __contains__(self, path):
        """Implement membership in terms of __getitem__"""
        try:
            self[path]
            return True
        except KeyError:
            return False

    def get(self, path, default=None):
        """Get some path or return default value"""
        try:
            return self[path]
        except KeyError:
            return default

    def source_for(self, path):
        """Proxy self.storage.source_for"""
        if isinstance(path, list):
            path = dot_joiner(path)
        return self.storage.source_for(self.prefixed_path_string(path))

    def __setitem__(self, path, value):
        """Set a key in the storage"""
        self.storage.add(self.prefixed_path_list([path]), value)

    def __delitem__(self, path):
        """Delete a key from the storage"""
        self.storage.delete(self.prefixed_path_string(path))

    def __iter__(self):
        """Iterate over the keys"""
        return iter(self.keys())

    def __len__(self):
        """Get number of keys we have"""
        return len(list(self.keys()))

    def __eq__(self, other):
        """Equal to another merged options if has same storage and prefix"""
        return isinstance(other, self.__class__) and other.storage is self.storage and other.prefix_list == self.prefix_list

    def values_for(self, path):
        """Get all known values for some path"""
        if isinstance(path, list):
            path = dot_joiner(path)
        path = self.prefixed_path_string(path)
        for info in self.storage.get_info(path):
            yield info.value_after(path)

    def prefixed(self, path):
        """Return a MergedOptions prefixed to this path"""
        if isinstance(path, six.string_types):
            path = [path]
        return self.__class__(self.prefixed_path_list(path), storage=self.storage)

    def prefixed_path_list(self, path):
        """Proxy the prefixed_path_list helper with prefix from this instance"""
        return prefixed_path_list(path, self.prefix_list)

    def prefixed_path_string(self, path):
        """Proxy the prefixed_path_string helper with prefix from this instance"""
        return prefixed_path_string(path, self.prefix_string)

    def keys(self):
        """Return a de-duplicated list of the keys we know about"""
        for key in self.storage.keys_after(self.prefix_string):
            prefixless = without_prefix(key, self.prefix_list)
            if prefixless:
                yield prefixless

    def items(self):
        """Iterate over [(key, value), ...] pairs"""
        for key in self.keys():
            yield key, self[key]

    def values(self):
        """Return the values"""
        for key in self.keys():
            yield self[key]

    def __repr__(self):
        return "MergedOptions({0})".format(self.prefix_string)




from option_merge.merge import ConverterProperty, KeyValuePairsConverter, AttributesConverter
from option_merge.merge import MergedOptions, BadPrefix
from option_merge.helper import NotFound

# Explicitly make these part of this module
NotFound = NotFound
BadPrefix = BadPrefix
MergedOptions = MergedOptions
ConverterProperty = ConverterProperty
AttributesConverter = AttributesConverter
KeyValuePairsConverter = KeyValuePairsConverter




from option_merge.helper import value_at, NotFound, dot_joiner, make_dict
from delfick_error import ProgrammerError
from collections import namedtuple
import six

class Path(namedtuple("Path", ("path", "data", "source"))):

    def keys_after(self, prefix):
        """All the keys after prefix"""
        parts = [part for part in self.path]
        for part in self.path:
            if prefix.startswith("{0}.".format(part)) or prefix == part:
                prefix = prefix[len(part)+1:]
                parts.pop(0)
            elif not prefix and parts:
                yield parts[0]
                return
            else:
                raise NotFound

        if parts:
            yield parts[0]
        elif isinstance(self.data, dict):
            for key in self.data.keys():
                yield key

    def value_after(self, prefix):
        """Returns the value after prefix"""
        parts = [part for part in self.path]
        for part in self.path:
            if prefix.startswith("{0}.".format(part)) or prefix == part:
                prefix = prefix[len(part)+1:]
                parts.pop(0)
            elif not prefix and parts:
                return make_dict(parts[0], parts[1:], self.data)
            else:
                raise NotFound

        if parts:
            return make_dict(parts[0], parts[1:], self.data)
        elif isinstance(self.data, dict):
            if prefix:
                if prefix in self.data:
                    return self.data[prefix]
                else:
                    raise NotFound

        return self.data

class Storage(object):
    """Holds the dataz"""

    def __init__(self):
        self.data = []
        self.deleted = []

    ########################
    ###   USAGE
    ########################

    def add(self, path, data, source=None):
        """Add data at the beginning"""
        if isinstance(path, six.string_types):
            raise ProgrammerError("Path should be a list\tgot={0}".format(type(path)))
        self.data.insert(0, (path, data, source))

    def get(self, path):
        """Get a single value from a path"""
        for info in self.get_info(path):
            return info.data
        raise KeyError(path)

    def source_for(self, path):
        """Find all the sources for a given path"""
        sources = []
        for info in self.get_info(path):
            if dot_joiner(info.path) == path and not isinstance(info.data, dict):
                if isinstance(info.source, list):
                    return [thing for thing in info.source]
                else:
                    return [info.source]
            else:
                if info.source not in sources:
                    sources.append(info.source)

        if sources:
            return sources
        else:
            raise KeyError(path)

    def delete(self, path):
        """Delete the first instance of some path"""
        for index, (info_path, data, _) in enumerate(self.data):
            dotted_info_path = dot_joiner(info_path)
            if dotted_info_path == path:
                del self.data[index]
                return
            elif dotted_info_path.startswith("{0}.".format(path)):
                del self.data[index]
                return
            elif not dotted_info_path or path.startswith("{0}.".format(dotted_info_path)):
                remainder = path
                if info_path:
                    remainder = path[len(dotted_info_path)+1:]
                if self.delete_from_data(data, remainder):
                    return

        raise KeyError(path)

    ########################
    ###   IMPLEMENTATION
    ########################

    def get_info(self, path):
        yielded = False
        if not self.data and not path:
            return

        for info_path, data, source in self.data:
            dotted_info_path = dot_joiner(info_path)
            if not info_path or not path or path == dotted_info_path or path.startswith("{0}.".format(dotted_info_path)) or dotted_info_path.startswith("{0}.".format(path)):
                try:
                    get_at = path
                    if info_path:
                        get_at = path[len(dotted_info_path)+1:]
                    found_path, val = value_at(data, get_at)
                    if hasattr(data, "source_for"):
                        source = data.source_for(get_at)
                    yield Path(info_path + found_path, val, source)
                    yielded = True
                except NotFound:
                    pass

        if not yielded:
            raise KeyError(path)

    def keys_after(self, path):
        """Get all the keys after this path"""
        done = set()
        stopped = set()
        for info in self.get_info(path):
            for key in info.keys_after(path):
                joined = dot_joiner([path, key])
                if not any(s == joined or joined.startswith("{0}.".format(s)) for s in stopped):
                    if key not in done:
                        yield key
                        done.add(key)

            if not isinstance(info.data, dict):
                stopped.add(dot_joiner(info.path))

    def delete_from_data(self, data, path):
        if not path or not isinstance(data, dict):
            return False

        keys = list(reversed(sorted(data.keys())))
        if path in keys:
            del data[path]
            return True

        for key in keys:
            if path.startswith("{0}.".format(key)):
                if self.delete_from_data(data[key], path[len(key)+1:]):
                    return True

